# -*- coding: utf-8 -*-
#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: embedsignature=True

cimport cython
import numpy as np

from cython.parallel import prange, threadid, parallel
from cythonarrays.numpy_types cimport *
from cythonarrays.array_shapes cimport ArrayShapes
from libc.math cimport exp


cdef class _Example(ArrayShapes):
    """
    Cython CDefClass for Example model
    with coordinates groups and zones
    """
    @cython.initializedcheck(False)
    cpdef char calc_model(self) except -1:
        """
        Calc the daily trips for all groups and zones
        """
        cdef char t
        cdef long32 g
        self.reset_array('trips_ij')
        with nogil, parallel(num_threads=self.n_threads):
            t = threadid()
            # loop over groups
            for g in prange(self.groups, schedule='guided'):
                with gil:
                    self.logger.info('calculate group {}'.format(g))
                # calc destination choice model for group g
                self._calc_p_destination(g)

    @cython.initializedcheck(False)
    cdef double _calc_weight_destination(self, double param,
                                         double minutes, double jobs) nogil:
        """calculate weight for destination"""
        cdef double weight = exp(param * minutes) * jobs
        return weight

    @cython.initializedcheck(False)
    cdef ARRAY_1D_d _calc_p_destination(self, long32 g) nogil:
        """Calc the destination choice probability for group g"""
        cdef double param, minutes, persons, jobs, weight, total_weight
        cdef long32 i, j
        cdef ARRAY_1D_d weights_j
        with gil:
            weights_j = np.empty((self.destinations), 'd')
        param = self._param_g[g]
        for i in range(self.origins):
            persons = self._persons_gi[g, i]
            total_weight = 0
            for j in range(self.destinations):
                jobs = self._jobs_j[j]
                minutes = self._km_ij[i, j]
                weight = self._calc_weight_destination(param, minutes, jobs)
                weights_j[j] = weight
                total_weight += weight
            factor = persons / total_weight
            for j in range(self.destinations):
                self._trips_ij[i, j] += factor * weights_j[j]
        return weights_j

    def calc_p_destination(self, g):
        """
        Calc the destination choice probability for group g

        Parameters
        ----------
        g : int
            the group number

        Returns
        -------
        weights : np.array
            array with the weights for each zone
        """
        return self._calc_p_destination(g)





"""
Defines Ctype names to be used in cython modules

Attributes
----------
numpy array dtype in cython

    BTYPE_t
    DTYPE_i1_t ... DTYPE_i8_t
    DTYPE_u1_t ... DTYPE_u8_t
    DTYPE_f_t
    DTYPE_d_t

Arrays/Memoryviews

    ARRAY_1D_d ... ARRAY_5D_d
    ARRAY_1D_f ... ARRAY_5D_f

    ARRAY_1D_i1 ... ARRAY_3D_i1
    ARRAY_1D_i1 ... ARRAY_3D_i1
    ARRAY_1D_i4 ... ARRAY_3D_i4
    ARRAY_1D_i8 ... ARRAY_3D_i8
    ARRAY_1D_u1 ... ARRAY_3D_u1
    ...
    ARRAY_1D_u8 ... ARRAY_3D_u8

"""

cimport cython
cimport numpy as np

ctypedef cython.int long32

ctypedef cython.uint ulong32

ctypedef cython.char BTYPE_t

ctypedef cython.longlong DTYPE_i8_t

ctypedef cython.ulonglong DTYPE_u8_t

ctypedef cython.double DTYPE_d_t

ctypedef cython.float DTYPE_f_t

ctypedef cython.char DTYPE_i1_t

ctypedef cython.uint DTYPE_u4_t

ctypedef cython.short DTYPE_i2_t

ctypedef cython.uchar DTYPE_u1_t

ctypedef cython.int DTYPE_i4_t

ctypedef cython.ushort DTYPE_u2_t

ctypedef cython.ssize_t Pointer

ctypedef cython.double[:, :, :, : , :] ARRAY_5D_d
ctypedef cython.double[:, :, :, :] ARRAY_4D_d
ctypedef cython.double[:, :, :] ARRAY_3D_d
ctypedef cython.double[:, :] ARRAY_2D_d
ctypedef cython.double[:] ARRAY_1D_d

ctypedef cython.float[:, :, :, : , :] ARRAY_5D_f
ctypedef cython.float[:, :, :, :] ARRAY_4D_f
ctypedef cython.float[:, :, :] ARRAY_3D_f
ctypedef cython.float[:, :] ARRAY_2D_f
ctypedef cython.float[:] ARRAY_1D_f

ctypedef cython.char[:] ARRAY_1D_b
ctypedef cython.char[:] ARRAY_1D_i1
ctypedef cython.short[:] ARRAY_1D_i2
ctypedef cython.int[:] ARRAY_1D_i4
ctypedef cython.longlong[:] ARRAY_1D_i8
ctypedef cython.ssize_t[:] ARRAY_1D_Ptrs

ctypedef cython.char[:, :] ARRAY_2D_b
ctypedef cython.char[:, :] ARRAY_2D_i1
ctypedef cython.short[:, :] ARRAY_2D_i2
ctypedef cython.int[:, :] ARRAY_2D_i4
ctypedef cython.longlong[:, :] ARRAY_2D_i8

ctypedef cython.char[:, :, :] ARRAY_3D_b
ctypedef cython.char[:, :, :] ARRAY_3D_i1
ctypedef cython.short[:, :, :] ARRAY_3D_i2
ctypedef cython.int[:, :, :] ARRAY_3D_i4
ctypedef cython.longlong[:, :, :] ARRAY_3D_i8

ctypedef cython.char[:, :, :, :] ARRAY_4D_i1
ctypedef cython.char[:, :, :, :, :] ARRAY_5D_i1
ctypedef cython.int[:, :, :, :] ARRAY_4D_i4

ctypedef cython.uchar[:] ARRAY_1D_u1
ctypedef cython.ushort[:] ARRAY_1D_u2
ctypedef cython.uint[:] ARRAY_1D_u4
ctypedef cython.ulonglong[:] ARRAY_1D_u8

ctypedef cython.uchar[:, :] ARRAY_2D_u1
ctypedef cython.ushort[:, :] ARRAY_2D_u2
ctypedef cython.uint[:, :] ARRAY_2D_u4
ctypedef cython.ulonglong[:, :] ARRAY_2D_u8

ctypedef cython.uchar[:, :, :] ARRAY_3D_u1
ctypedef cython.ushort[:, :, :] ARRAY_3D_u2
ctypedef cython.uint[:, :, :] ARRAY_3D_u4
ctypedef cython.ulonglong[:, :, :] ARRAY_3D_u8

ctypedef fused np_signed_int:
    cython.char
    cython.short
    cython.int
    cython.longlong

ctypedef fused np_int_4_8:
    cython.int
    cython.longlong

ctypedef fused np_uint:
    cython.uchar
    cython.ushort
    cython.uint
    cython.ulonglong

ctypedef fused np_floating:
    cython.float
    cython.double

ctypedef fused np_integer:
    cython.char
    cython.short
    cython.int
    cython.longlong
    cython.uchar
    cython.ushort
    cython.uint
    cython.ulonglong

ctypedef fused np_numeric:
    cython.char
    cython.short
    cython.int
    cython.longlong
    cython.uchar
    cython.ushort
    cython.uint
    cython.ulonglong
    cython.float
    cython.double


ctypedef np_signed_int[:] ARRAY_1D_i
ctypedef np_int_4_8[:] ARRAY_1D_int
ctypedef np_uint[:] ARRAY_1D_u
ctypedef np_integer[:] ARRAY_1D_iu
ctypedef np_floating[:] ARRAY_1D_fd
ctypedef np_numeric[:] ARRAY_1D_n

ctypedef np_signed_int[:, :] ARRAY_2D_i
ctypedef np_int_4_8[:, :] ARRAY_2D_int
ctypedef np_uint[:, :] ARRAY_2D_u
ctypedef np_integer[:, :] ARRAY_2D_iu
ctypedef np_floating[:, :] ARRAY_2D_fd
ctypedef np_numeric[:, :] ARRAY_2D_n

ctypedef np_signed_int[:, :, :] ARRAY_3D_i
ctypedef np_int_4_8[:, :, :] ARRAY_3D_int
ctypedef np_uint[:, :, :] ARRAY_3D_u
ctypedef np_integer[:, :, :] ARRAY_3D_iu
ctypedef np_floating[:, :, :] ARRAY_3D_fd
ctypedef np_numeric[:, :, :] ARRAY_3D_n

ctypedef np_signed_int[:, :, :, :] ARRAY_4D_i
ctypedef np_int_4_8[:, :, :, :] ARRAY_4D_int
ctypedef np_uint[:, :, :, :] ARRAY_4D_u
ctypedef np_integer[:, :, :, :] ARRAY_4D_iu
ctypedef np_floating[:, :, :, :] ARRAY_4D_fd
ctypedef np_numeric[:, :, :, :] ARRAY_4D_n

ctypedef np_signed_int[:, :, :, :, :] ARRAY_5D_i
ctypedef np_int_4_8[:, :, :, :, :] ARRAY_5D_int
ctypedef np_uint[:, :, :, :, :] ARRAY_5D_u
ctypedef np_integer[:, :, :, :, :] ARRAY_5D_iu
ctypedef np_floating[:, :, :, :, :] ARRAY_5D_fd
ctypedef np_numeric[:, :, :, :, :] ARRAY_5D_n



import logging
import datetime
import os
import inspect


class SimLogger(object):
    """
    Singleton Class for the logging
    """
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(SimLogger, cls).__new__(
                                cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        """ init the sim-logger"""
        if not hasattr(self, 'packages'):
            self.packages = set()
        if not hasattr(self, 'filehandler_initialized'):
            self.filehandler_initialized = False
        # add at least a console handler
        if not hasattr(self, 'ch'):
            self.add_console_handler()

    def add_package(self, package):
        """add single package to the list of logged packages"""
        if package not in self.packages:
            self.packages.add(package)
            self.create_logger_for_package(package)

    def add_packages(self, packages):
        """add list of packages to the list of logged packages"""
        for package in packages:
            self.add_package(package)

    def configure(self,
                  LOG_FOLDER,
                  scenario,
                  mode='a'):
        """
        configures the file handlers for the logger

        Parameters
        ----------
        LOG_FOLDER : str
            the path to the log-folder
        scenario : str
            the name of the scenario used in the log file
        mode : str, optional(Default: 'a')
            a: append
            w: recreate logfile
        """

        iso_date = datetime.date.today().isoformat()
        # define the file where to log
        LOGGING_FILE = os.path.join(LOG_FOLDER,
                                    '{sc}_{date}.log'.format(sc=scenario,
                                                             date=iso_date))
        LOGGING_FILE_DEBUG = os.path.join(LOG_FOLDER,
                                          '{sc}_{date}_debug.log'.format(
                                              sc=scenario,
                                              date=iso_date))

        # create file handler which logs only info messages
        self.fh = logging.FileHandler(LOGGING_FILE, mode=mode)
        self.fh.setLevel(logging.INFO)

        # create file handler which logs even debug messages
        self.fh_debug = logging.FileHandler(LOGGING_FILE_DEBUG, mode=mode)
        self.fh_debug.setLevel(logging.DEBUG)

        # create file formatter and add it to the handlers
        self.fh_debug.setFormatter(self.get_debug_formatter())
        self.fh.setFormatter(self.get_info_formatter())
        self.filehandler_initialized = True

        # configure the file handlers to all given packages
        for package in self.packages:
            self.create_logger_for_package(package)

    def add_console_handler(self):
        # create console handler with a higher log level
        self.ch = logging.StreamHandler()
        self.ch.setLevel(logging.INFO)
        self.ch.setFormatter(self.get_info_formatter())

    def get_debug_formatter(self):
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        return formatter

    def get_info_formatter(self):
        msg = '%(asctime)s.%(msecs).03d-%(levelname)s->%(message)s'
        formatter_info = logging.Formatter(fmt=msg, datefmt='%H:%M:%S')
        return formatter_info

    def create_logger_for_package(self, package):
        # create a logger for the package
        logger = logging.getLogger(package)
        logger.setLevel(logging.DEBUG)

        # remove existing handlers
        del logger.handlers[:]
        logger.addHandler(self.ch)

        if self.filehandler_initialized:
            # add the handlers to the logger
            logger.addHandler(self.fh)
            logger.addHandler(self.fh_debug)

    @staticmethod
    def get_caller_name(skip=2):
        """
        Get a name of a caller in the format module.class.method

        `skip` specifies how many levels of stack to skip while getting caller
        name. skip=1 means "who calls me", skip=2 "who calls my caller" etc.

        An empty string is returned if skipped levels exceed stack height
        """
        stack = inspect.stack()
        start = 0 + skip
        if len(stack) < start + 1:
            return ''
        parentframe = stack[start][0]

        name = []
        module = inspect.getmodule(parentframe)
        # `modname` can be None when frame is executed directly in console
        # TODO(techtonik): consider using __main__
        if module:
            name.append(module.__name__)
        # detect classname
        if 'self' in parentframe.f_locals:
            # I don't know any way to detect call from the object method
            # XXX: there seems to be no way to detect static method call
            #   - it will be just a function call
            name.append(parentframe.f_locals['self'].__class__.__name__)
        codename = parentframe.f_code.co_name
        # top level usually
        if codename != '<module>' and codename != '__init__':
            name.append(codename)  # function or a method
        del parentframe
        return ".".join(name)

    def get(self, instance):
        """
        returns a logger for the current package

        Parameters
        ----------
        instance : object
            the instance to inspect
        """
        if hasattr(instance, '__module__'):
            package = instance.__module__.split('.')[0]
        else:
            package = None
        if package in self.packages:
            caller_name = '.'.join([instance.__module__,
                                    instance.__class__.__name__])
        else:
            caller_name = self.get_caller_name(skip=3)
            print('''logging uses inspect-module, may be slow.
            Consider import from fully qualified namespace package
            (from namespace.{package} import xyz
            instead of from {package} import xyz)'''.format(package=package))

        logger = logging.getLogger(caller_name)
        if not logger.handlers:
            logger.addHandler(logging.NullHandler())
        return logger


def get_module_logger(module_name):
    """create Module logger named after the module"""
    module_logger = logging.getLogger(module_name)
    # Add a NullHandler for the case if no logging is configured by the
    # Application
    module_logger.addHandler(logging.NullHandler())
    return module_logger


def get_logger(instance):
    """
    for the instance, get the package,
    add the package to the list of loggers,
    and return the logger for the instance

    Parameters
    ----------
    instance : object

    Returns
    -------
    logger : logger-instace

    """
    sim_logger = SimLogger()
    if hasattr(instance, '__module__'):
        sim_logger.add_package(instance.__module__.split('.')[0])
    return sim_logger.get(instance)



# -*- coding: utf-8 -*-

import numpy as np
from numpy.testing import assert_array_equal


class ArrayDescriptor(object):
    """
    describes an array used as an instance attribute of a cython class

    Parameters
    ----------
    name : str
        the name of the array
    dtype : str
        the numpy-dtype of the array for internal storage
    ndim : int
        the number of dimensions of the array
    shape : tuple of str or int
        the shape of the array
    dtype_numpy : str
        the numpy-dtype of the array that should be returned
    default : number
        the default value
    """
    def __init__(self, name, dtype, ndim,
                 shape=None, dtype_numpy=None, default=None):
        self.name = name
        self.dtype = dtype
        self.ndim = ndim
        self.shape = shape
        self._dtype_numpy = dtype_numpy
        self.default = default

    def __repr__(self):
        txt = 'array {0.name} of dtype {0.dtype}, ndim {0.ndim}'
        if self.shape is not None:
            txt += ' and shape {0.shape}'
        return txt.format(self)

    @staticmethod
    def to_dim(arr, ndim):
        """
        Casts the array to ndim dimensions

        Parameters
        ----------
        ndim : integer
            if self.ndim > ndim and the shape of the first axis <> 0:
            raise ValueError

        Returns
        -------
        arr : array of ndim dimensions
        """

        dim_diff = ndim - arr.ndim
        if dim_diff > 0:
            return arr.__getitem__((np.newaxis, ) * dim_diff)
        elif dim_diff < 0:
            dim_diff *= -1
            if arr.shape[:dim_diff] == tuple([1] * dim_diff):
                new_shape = arr.shape[dim_diff:]
                return arr.reshape(new_shape)
            else:
                msg = f'cannot reshape shape {arr.shape} to {ndim} dimensions'
                raise ValueError(msg)
        else:
            return arr

    def validate_array(self, arr, instance=None):
        """
        checks if the ndim (and the shape, if specified) match
        and casts the array to the dtype specified
        """
        # cast arr to numpy array
        if not isinstance(arr, np.ndarray):
                arr = np.array(arr)
        msg = f'{self.name}: ndim target: {self.ndim}, actual: {arr.ndim}'
        # if ndim does not match
        try:
            assert self.ndim == arr.ndim, msg
        except AssertionError as err:
            # try to add additional dimensions
            try:
                arr = self.to_dim(arr, self.ndim)
            except ValueError:
                raise err
        if self.shape is not None:
            shape = self.get_shape(instance)
            msg = f'{self.name}: shape target: {shape}, actual: {arr.shape}'
            assert_array_equal(arr.shape, shape, msg)

        # convert bool array to i1 with view instead of astype
        if issubclass(arr.dtype.type, np.bool_):
            new_type = np.dtype(self.dtype).type
            if issubclass(new_type, np.int8):
                return arr.view(dtype='i1')
            elif issubclass(new_type, np.uint8):
                return arr.view(dtype='u1')

        return arr.astype(self.dtype, copy=False)

    def get_shape(self, instance):
        shape = []
        for d in range(self.ndim):
            if isinstance(self.shape[d], str):
                v = getattr(instance, self.shape[d])
            else:
                v = self.shape[d]
            shape.append(v)
        return shape

    @property
    def dtype_numpy(self):
        if self._dtype_numpy is None:
            return self.dtype
        else:
            return self._dtype_numpy

#     do not modify the dtype later
#     @dtype_numpy.setter
#     def dtype_numpy(self, value):
#         self._dtype_numpy = value

    @property
    def shape(self):
        return self._shape

    @shape.setter
    def shape(self, value):
        """convert the shape to a tuple"""
        if isinstance(value, (np.int, np.long)):
            value = (value, )
        elif isinstance(value, (str, bytes)):
            splitted = value.split(',')
            value = []
            for s in splitted:
                # strip whitespace
                st = s.strip()
                try:
                    value.append(int(st))
                except ValueError:
                    value.append(st)
        self._shape = value



# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

from cythonarrays.configure_logger import (SimLogger,
                                           get_module_logger,
                                           get_logger)
import tempfile
import shutil
import os
import glob


class Test01_Logger:
    """Test the the SimLogger"""
    @classmethod
    def setup_class(cls):
        cls.LOG_FOLDER = tempfile.mkdtemp(prefix='Log_')
        print(cls.LOG_FOLDER)
        cls.scenario = 'TestScenario'
        sim_logger = SimLogger()
        print(sim_logger.packages)
        sim_logger.add_packages(['cythonarrays',
                                 'cythoninstallhelpers',
                                 'numpy'])
        sim_logger.configure(LOG_FOLDER=cls.LOG_FOLDER,
                             scenario=cls.scenario)
        print(sim_logger.packages)

    @classmethod
    def teardown_class(cls):
        try:
            shutil.rmtree(cls.LOG_FOLDER)
        except PermissionError:
            pass

    def test_02_log_something(self):
        """log something to the logger"""
        logger = get_logger(self)
        logger.info('Info in test_02')
        logger.debug('Debug in test_02')
        logger.warning('Warn in test_02')
        logfiles = glob.glob(os.path.join(self.LOG_FOLDER,
                                          '{}*.log'.format(self.scenario)))
        assert logfiles
        print(logfiles)
        for logfile in logfiles:
            with open(logfile) as f:
                for line in f:
                    print(line.strip())



    def test_03_module_logger(self):
        """Test the module logger"""
        module_logger = get_module_logger(self.__module__)
        module_logger.info('Info in Module {}'.format(self.__module__))


if __name__ == '__main__':
    pytest.main()


# -*- coding: utf-8 -*-
"""
Created on Fri Jun 10 21:00:21 2016

@author: MaxBohnet
"""

import tempfile
import numpy as np
import pytest
import xarray as xr

from cythonarrays.tests.example_python import Example
from cythonarrays.tests.simple_python import Simple
import pyximport; pyximport.install()
from .example_cython import (_Example)


@pytest.fixture(scope='class')
def km_ij(request):
    """
    travel times
    """
    array = np.array([[1, 4, 5],
                      [2, 1, 7],
                      [5, 7, 2]])
    return array


@pytest.fixture(scope='class')
def jobs(request):
    """
    jobs
    """
    jobs = np.array([100, 200, 300])
    return jobs


@pytest.fixture(scope='class')
def params_groups(request):
    """
    different values for saving categories
    """
    return np.array([-0.2, -0.1])


@pytest.fixture(scope='class')
def persons_gi(request):
    """
    different values for home zone
    """
    arr = np.array([[100, 0, 200],
                    [0, 250, 50]])
    return arr


@pytest.fixture(scope='class')
def example(km_ij, jobs, params_groups, persons_gi):
    """
    different values for home zone
    """
    groups, zones = persons_gi.shape
    example = Example(groups, zones)
    example.set_array('km_ij', km_ij)
    example.set_array('jobs_j', jobs)
    example.set_array('param_g', params_groups)
    example.set_array('persons_gi', persons_gi)
    return example


@pytest.fixture()
def tempfile_h5():
    return tempfile.mktemp(suffix='h5')


class Test01_ExampleCDefClass:
    """Test the Example CDefClass"""

    def test_01_test_init_array(self, persons_gi):
        """Test the Example CDefClass creation"""
        groups, zones = persons_gi.shape
        example = Example(groups, zones)

        # set correct shape
        values = np.arange(zones)
        example.set_array('jobs_j', np.arange((zones)))
        np.testing.assert_array_equal(example.jobs_j, values)
        np.testing.assert_array_equal(example._jobs_j, values)

        # init the array with a default value
        example.init_array('jobs_j', default=0)
        np.testing.assert_array_equal(example.jobs_j, np.zeros(zones))

        # init the array with the other default value
        example.init_array('jobs_j', default=2)
        np.testing.assert_array_equal(example.jobs_j,
                                      np.full(zones, 2, dtype='d'))

        # new default value will be stored
        example.reset_array('jobs_j')
        np.testing.assert_array_equal(example.jobs_j,
                                      np.full(zones, 2, dtype='d'))

        # set new shape, keep old default value
        example.init_array('jobs_j', shape='groups')
        np.testing.assert_array_equal(example.jobs_j,
                                      np.full(groups, 2, dtype='d'))

    def test_01a_test_init_array_with_other_datatypes(self, persons_gi):
        """Test the Example CDefClass creation"""
        example = Example(groups=2, origins=3)
        # test setting with a range
        example.jobs_j = range(3)
        np.testing.assert_array_equal(example.jobs_j, np.array([0.0, 1.0, 2.0]))
        # test setting with a list
        example.jobs_j = list(range(3))
        np.testing.assert_array_equal(example.jobs_j, np.array([0.0, 1.0, 2.0]))
        #  test setting array with a tuple
        example.jobs_j = tuple(range(3))
        np.testing.assert_array_equal(example.jobs_j, np.array([0.0, 1.0, 2.0]))
        #  test setting array with a list of random objects
        pattern = r"""float\(\) argument must be a string or a number, not .*"""
        with pytest.raises(TypeError, match=pattern):
            example.jobs_j = [float, type, int]


    def test_02_test_shape(self, persons_gi):
        """
        Test if the shapes of array are controlled correctly
        """
        groups, zones = persons_gi.shape
        example = Example(groups, zones)

        # jobs_j should have the shape (zones)

        # correct shape
        arr = np.ones((zones))
        example.jobs_j = arr

        # try to set values with the wrong shape
        arr = np.ones((groups))
        pattern = r"""
Arrays are not equal
jobs_j: shape target: \[3\], actual: \(2,\)
"""
        with pytest.raises(AssertionError,
                           match=pattern):
            example.jobs_j = arr

        # change zones and try again
        example.destinations = groups
        # now the array fits to the shape defined
        example.jobs_j = arr

    def test_03_test_dimensions(self, persons_gi):
        """Test the dimensions"""
        groups, zones = persons_gi.shape
        example = Example(groups, zones)

        # try to set shape with the wrong number of dimensions
        pattern = r"1 Dimensions required, shape \[\'groups\', \'zones\'\] has 2 dimensions"
        with pytest.raises(ValueError,
                           match=pattern):
            example.init_array('jobs_j', shape='groups, zones')

        # access a non-initialized array
        target = np.empty((0, 0))
        actual = example.not_initialized_ij
        np.testing.assert_array_equal(actual, target)

        # set arbitrary values to non-initialized 2D-Array
        a = np.array([2, 3, 4], dtype='i4')
        # with less dimensions than required ...
        example.not_initialized_ij = a
        b = example.not_initialized_ij
        # another dimension is added
        assert b.ndim == 2
        assert b.shape == (1, 3)
        print(example.not_initialized_ij)

        # change value in cython-memoryview
        example._not_initialized_ij[0, 1] = 99
        target = np.array([[2, 99, 4]], dtype='i4')
        # the value will be changed in the numpy-view of the array
        np.testing.assert_array_equal(example.not_initialized_ij, target)
        # as this is a view on the input-data, it will be changed there, too
        np.testing.assert_array_equal(a[1], 99)

        # the right number of dimensions should be fine
        c = np.array([[0, 1],
                      [2, 3]], dtype='i4')
        example.not_initialized_ij = c
        d = example.not_initialized_ij
        np.testing.assert_array_equal(c, d)

        # too many dimensions will raise an error
        e = np.array([[[0, 1],
                       [2, 3]],
                      [[4, 5],
                       [6, 7]]], dtype='i4')
        with pytest.raises(AssertionError,
                           match="not_initialized_ij: ndim target: 2, actual: 3"):
            example.not_initialized_ij = e

        # when the dimensions can be reduced to the target dimensions,
        # it should be fine
        f = np.array([[[0, 1],
                       [2, 3]]], dtype='i4')
        example.not_initialized_ij = f
        g = example.not_initialized_ij
        np.testing.assert_array_equal(g, f[0])

        # change value in cython-memoryview
        example._not_initialized_ij[1, 1] = 88
        target = np.array([[0, 1],
                           [2, 88]], dtype='i4')
        # the value will be changed in the numpy-view of the array
        np.testing.assert_array_equal(example.not_initialized_ij, target)
        # as this is a view on the input-data, it will be changed there, too
        np.testing.assert_array_equal(f[0, 1, 1], 88)

        # use more dimensions than required
        # reset the array first
        del example.not_initialized_ij
        print(example.not_initialized_ij)
        c = np.array([[[2, 3, 4]]], dtype='i4')
        example.not_initialized_ij = c
        b = example.not_initialized_ij
        # one dimension can be taken away
        assert b.ndim == 2
        assert b.shape == (1, 3)
        print(example.not_initialized_ij)

        # use the correct number of dimensions
        # reset the array first
        del example.not_initialized_ij
        print(example.not_initialized_ij)
        d = np.array([[2, 3, 4]], dtype='i4')
        example.not_initialized_ij = d
        b = example.not_initialized_ij
        # one dimension can be taken away
        assert b.ndim == 2
        assert b.shape == (1, 3)
        print(example.not_initialized_ij)

    def test_04_test_dtype(self, persons_gi):
        """Test the dimensions"""
        groups, zones = persons_gi.shape
        example = Example(groups, zones)

        # dtype of jobs_j: f8
        # dtype of not_initialized_ij: i4
        # dtype of n_threads: char

        # test if n_threads is really a signed char
        example.n_threads = 127
        assert example.n_threads == 127

        with pytest.raises(OverflowError,
                           match='value too large to convert to char'):
            example.n_threads = 128

        # test if jobs_j is really an array of double precision
        arr_i4 = np.array([2, 3, 4], dtype='i4')
        example.jobs_j = arr_i4
        assert example.jobs_j.dtype == np.dtype('f8')
        # because the array set to jobs_j is of different dtype,
        # it had to be converted to double. So no view is used,
        # but a copy of the data
        arr_i4[1] = -1
        # the data in jobs_j stays untouched
        assert example.jobs_j[1] == 3

        # test if not_initialized_ij is really an array of 32bit integer
        arr_i4 = np.array([[2, 3, 4], [5, 6, 7]], dtype='i4')
        example.not_initialized_ij = arr_i4
        assert example.not_initialized_ij.dtype == np.dtype('i4')
        # because the array set to not_initialized_ij is of the same dtype,
        # a view on the data could have been used
        arr_i4[1, 2] = -1
        # the data in not_initialized_ij changes
        assert example.not_initialized_ij[1, 2] == -1

    def test_10_test_model(self, example):
        """Test the Example CDefClass model"""
        res = example.calc_model()
        print(example.trips_ij)
        total_trips_target = example.persons_gi.sum()
        total_trips_actual = example.trips_ij.sum()
        np.testing.assert_almost_equal(total_trips_target, total_trips_actual)

    def test_11_dataset(self, example):
        """Test the creation of an xarrays Dataset linked to the Cythonclass"""
        example.zonenumbers_i = np.array([100, 200, 300])
        example.groupnames_g = np.array(['Female', 'Male'], dtype='O')
        example.create_ds()
        assert isinstance(example.ds, xr.Dataset)
        for attr, dtype in example.dtypes.items():
            data_array = example.ds[attr]
            # test if the shapes are correct
            if dtype.shape:
                np.testing.assert_array_equal(dtype.get_shape(example),
                                              data_array.shape,
                                              'shape not correct')
            else:
                # not initialized array
                assert not np.any(data_array.shape), 'shape not initialized'
            #test if the datatypes are correct
            assert np.dtype(dtype.dtype) == data_array.dtype,  'dtype not correct'
        print(example.ds)

    def test_20_save_and_read_ds(self, example, tempfile_h5):
        """Test that saving and re-reading the data works"""
        example.zonenumbers_i = np.array([100, 200, 300])
        example.groupnames_g = np.array(['Female', 'Male'], dtype='O')
        example.create_ds()
        example.save_dataset_to_netcdf(tempfile_h5)

        # read the data into new class
        new_example = Example.from_netcdf(tempfile_h5)
        # assert that the data match
        np.testing.assert_array_equal(example.jobs_j, new_example.jobs_j)
        np.testing.assert_array_equal(example.groupnames_g,
                                      new_example.groupnames_g)

        # assert, that the array are correctly linked to the Dataset
        new_example.calc_model()
        old_value = new_example.ds.trips_ij.sum()
        # calculate the model with 77 persons more
        new_example.persons_gi[1, 0] += 77
        new_example.calc_model()
        new_value = new_example.ds.trips_ij.sum()
        np.testing.assert_allclose(old_value + 77, new_value)

        print(new_example.ds)

    def test_21_assign_bool(self, example):
        """Test boolean array"""
        # the boolean arrays are initialized with 0 and 1
        np.testing.assert_array_equal(example.valid_g.astype(bool),
                                      ~example.invalid_g.astype(bool))
        example.valid_g = np.zeros((example.groups), dtype=bool)
        example.invalid_g = np.ones((example.groups), dtype=bool)
        np.testing.assert_array_equal(example.valid_g.astype(bool),
                                      ~example.invalid_g.astype(bool))
        print(example.valid_g)
        print(example.invalid_g)

    def test_30_test_init_array(self, persons_gi):
        """Test the Example CDefClass creation"""
        example = Example(groups=7, origins=5)

        example.init_array('param_g', 7)
        assert example.param_g.shape == (7, )

    def test_32_test_not_init_array(self):
        """Test the Example CDefClass creation"""
        example = Example(groups=7, origins=5,
                          init_arrays=False, threading=False)
        with pytest.raises(AttributeError):
            print(example.not_initialized_ij)

    def test_32_test_not_init_array(self):
        """Test the Example CDefClass creation"""
        example = Example(groups=7, origins=5,
                          init_arrays=False, threading=False)
        with pytest.raises(AttributeError):
            print(example.not_initialized_ij)

    def test_33_del_property(self):
        """Test the Example CDefClass creation"""
        example = Example(groups=7, origins=5)
        print(example.km_ij)
        # remove the shape
        example.dtypes['km_ij'].shape = None

        # reset to default (not defined)
        example.reset_array('km_ij')
        np.testing.assert_array_equal(example.km_ij, np.NAN)

        # reset to default (new value defined)
        example.dtypes['km_ij'].default = 11
        example.reset_array('km_ij')
        np.testing.assert_array_equal(example.km_ij, 11)

        # delete values
        del example.km_ij
        np.testing.assert_array_equal(example.km_ij.shape, (0, 0))

        # set new values
        new_array = np.arange(10).reshape(2, 5)
        example.set_array('km_ij', value=new_array, shape=(2, 5))
        np.testing.assert_array_equal(example.km_ij, new_array)

    def test_34_save_ds(self, tempfile_h5):
        """Test the Example CDefClass creation"""
        example = Example(groups=7, origins=5, )
        example.save_dataset_to_netcdf(tempfile_h5)


class Test02_SimpleCDefClass:
    """Test the Simple CDefClass"""
    def test_01_simple_class(self):
        """Test some features of the CDef-Class"""
        simple = Simple(2, 3, init_arrays=True)
        np.testing.assert_array_equal(simple.param_g, [-0.1, -0.1])
        list_persons_gi = [[3, 4, 5], [6, 7, 8]]
        simple.persons_gi = list_persons_gi
        np.testing.assert_array_equal(simple.persons_gi, np.array(list_persons_gi))

        # add coordinates to the axis
        simple.zonenumbers_i = [10, 20, 30]
        simple.groupnames_g = ['G1', 'G2']

        #  test that access with coordinates work
        simple.param_g = [-0.1, 0.2]
        simple.create_ds()
        print(simple.ds)
        np.testing.assert_array_equal(simple.ds.param_g.loc['G1'], -0.1)
        np.testing.assert_array_equal(simple.ds.param_g.loc['G2'], 0.2)

        np.testing.assert_array_equal(simple.ds.persons_gi.loc['G1', 30], 5)
        np.testing.assert_array_equal(simple.ds.persons_gi.loc['G2', 20], 7)


class Test04_Test_Instantiation:
    """Test the instantiation of the class"""

    def test041_test_instantiation(self):
        """
        not subclassing shoud raise a NotImplementedError
        because the dtype is not initialized"""
        with pytest.raises(NotImplementedError) as e:
            example = _Example()
        print(e.value)

    def test042_test_nan(self, persons_gi):
        """Test nan"""
        groups, zones = persons_gi.shape
        example = Example(groups, zones)

        assert example.isnan_py(np.NAN)
        assert not example.isnan_py(np.NINF)
        assert not example.isnan_py(np.Inf)
        assert not example.isnan_py(0)
        assert not example.isnan_py(1)
        assert not example.isnan_py(-1)


if __name__ == '__main__':
    pytest.main()